@@ -1,5 +1,3 @@
-#if MOLD_ARM32 || MOLD_ARM64 || MOLD_PPC32 || MOLD_PPC64V1 || MOLD_PPC64V2
-
 // RISC instructions are usually up to 4 bytes long, so the immediates of
 // their branch instructions are naturally smaller than 32 bits.  This is
 // contrary to x86-64 on which branch instructions take 4 bytes immediates
@@ -22,6 +20,9 @@
 // we don't need to try too hard to reduce thunk size to the absolute
 // minimum.
 
+#if MOLD_ARM32 || MOLD_ARM64 || MOLD_PPC32 || MOLD_PPC64V1 || MOLD_PPC64V2 || \
+    MOLD_LOONGARCH64 || MOLD_LOONGARCH32
+
 #include "mold.h"
 
 #include <tbb/parallel_for.h>
@@ -29,15 +30,18 @@
 
 namespace mold::elf {
 
+using E = MOLD_TARGET;
+
 // Returns a branch reach in bytes for a given target.
-template <typename E>
 static consteval i64 max_distance() {
   // ARM64's branch has 26 bits immediate. The immediate is padded with
   // implicit two-bit zeros because all instructions are 4 bytes aligned
   // and therefore the least two bits are always zero. So the branch
   // operand is effectively 28 bits long. That means the branch range is
   // [-2^27, 2^27) or PC ± 128 MiB.
-  if (is_arm64<E>)
+  //
+  // LoongArch's BR instruction also takes a 26 bit immediate.
+  if (is_arm64<E> || is_loongarch<E>)
     return 1 << 27;
 
   // ARM32's Thumb branch has 24 bits immediate, and the instructions are
@@ -57,30 +61,12 @@ static consteval i64 max_distance() {
 
 // We create thunks for each 12.8/1.6/3.2 MiB code block for
 // ARM64/ARM32/PPC, respectively.
-template <typename E>
-static constexpr i64 batch_size = max_distance<E>() / 10;
+static constexpr i64 batch_size = max_distance() / 10;
 
 // We assume that a single thunk group is smaller than 100 KiB.
 static constexpr i64 max_thunk_size = 102400;
 
-// Returns true if a given relocation is of type used for function calls.
-template <typename E>
-static bool needs_thunk_rel(const ElfRel<E> &r) {
-  u32 ty = r.r_type;
-
-  if constexpr (is_arm64<E>) {
-    return ty == R_AARCH64_JUMP26 || ty == R_AARCH64_CALL26;
-  } else if constexpr (is_arm32<E>) {
-    return ty == R_ARM_JUMP24 || ty == R_ARM_THM_JUMP24 ||
-           ty == R_ARM_CALL   || ty == R_ARM_THM_CALL   ||
-           ty == R_ARM_PLT32;
-  } else if constexpr (is_ppc32<E>) {
-    return ty == R_PPC_REL24  || ty == R_PPC_PLTREL24 || ty == R_PPC_LOCAL24PC;
-  } else {
-    static_assert(is_ppc64<E>);
-    return ty == R_PPC64_REL24 || ty == R_PPC64_REL24_NOTOC;
-  }
-}
+static_assert(max_thunk_size / E::thunk_size < INT16_MAX);
 
 template <typename E>
 static bool is_reachable(Context<E> &ctx, InputSection<E> &isec,
@@ -116,7 +102,7 @@ static bool is_reachable(Context<E> &ctx, InputSection<E> &isec,
   // PowerPC before Power9 lacks PC-relative load/store instructions.
   // Functions compiled for Power9 or earlier assume that r2 points to
   // GOT+0x8000, while those for Power10 uses r2 as a scratch register.
-  // We need to a thunk to reconstruct r2 for interworking.
+  // We need to a thunk to recompute r2 for interworking.
   if constexpr (is_ppc64v2<E>) {
     if (rel.r_type == R_PPC64_REL24 && !sym.esym().preserves_r2())
       return false;
@@ -130,10 +116,9 @@ static bool is_reachable(Context<E> &ctx, InputSection<E> &isec,
   i64 A = get_addend(isec, rel);
   i64 P = isec.get_addr() + rel.r_offset;
   i64 val = S + A - P;
-  return -max_distance<E>() <= val && val < max_distance<E>();
+  return -max_distance() <= val && val < max_distance();
 }
 
-template <typename E>
 static void reset_thunk(RangeExtensionThunk<E> &thunk) {
   for (Symbol<E> *sym : thunk.symbols) {
     sym->extra.thunk_idx = -1;
@@ -143,16 +128,15 @@ static void reset_thunk(RangeExtensionThunk<E> &thunk) {
 }
 
 // Scan relocations to collect symbols that need thunks.
-template <typename E>
 static void scan_rels(Context<E> &ctx, InputSection<E> &isec,
-                      RangeExtensionThunk<E> &thunk) {
+                      RangeExtensionThunk<E> &thunk, i64 thunk_idx) {
   std::span<const ElfRel<E>> rels = isec.get_rels(ctx);
   std::vector<RangeExtensionRef> &range_extn = isec.extra.range_extn;
   range_extn.resize(rels.size());
 
   for (i64 i = 0; i < rels.size(); i++) {
     const ElfRel<E> &rel = rels[i];
-    if (!needs_thunk_rel(rel))
+    if (!is_func_call_rel(rel))
       continue;
 
     // Skip if the symbol is undefined. apply_reloc() will report an error.
@@ -174,7 +158,7 @@ static void scan_rels(Context<E> &ctx, InputSection<E> &isec,
 
     // Otherwise, add the symbol to the current thunk if it's not
     // added already.
-    range_extn[i].thunk_idx = thunk.thunk_idx;
+    range_extn[i].thunk_idx = thunk_idx;
     range_extn[i].sym_idx = -1;
 
     if (sym.flags.exchange(-1) == 0) {
@@ -184,51 +168,58 @@ static void scan_rels(Context<E> &ctx, InputSection<E> &isec,
   }
 }
 
-template <typename E>
-void create_range_extension_thunks(Context<E> &ctx, OutputSection<E> &osec) {
-  std::span<InputSection<E> *> m = osec.members;
+template <>
+void OutputSection<E>::create_range_extension_thunks(Context<E> &ctx) {
+  using Thunk = RangeExtensionThunk<E>;
+
+  std::span<InputSection<E> *> m = members;
   if (m.empty())
     return;
 
-  m[0]->offset = 0;
-
   // Initialize input sections with a dummy offset so that we can
   // distinguish sections that have got an address with the one who
   // haven't.
-  tbb::parallel_for((i64)1, (i64)m.size(), [&](i64 i) {
-    m[i]->offset = -1;
+  tbb::parallel_for_each(m, [](InputSection<E> *isec) {
+    isec->offset = -1;
   });
 
   // We create thunks from the beginning of the section to the end.
   // We manage progress using four offsets which increase monotonically.
   // The locations they point to are always A <= B <= C <= D.
   //
+  // Input sections between B and C are in the current batch.
+  //
   // A is the input section with the smallest address than can reach
-  // anywhere from the current batch.
+  // from the current batch.
   //
   // D is the input section with the largest address such that the thunk
-  // is reachable from the current batch if it's inserted right before D
-  // but at least partially unreacahble if it's inserted after D.
+  // is reachable from the current batch if it's inserted right before D.
   //
   //  ................................ <input sections> ............
   //     A    B    C    D
   //                    ^ We insert a thunk for the current batch just before D
-  //          <--->       The current batch, which is smaller than batch_size
-  //     <-------->       Smaller than max_distance
-  //          <-------->  Smaller than max_distance
+  //          <--->       The current batch, which is smaller than BATCH_SIZE
+  //     <-------->       Smaller than MAX_DISTANCE
+  //          <-------->  Smaller than MAX_DISTANCE
   //     <------------->  Reachable from the current batch
   i64 a = 0;
   i64 b = 0;
   i64 c = 0;
   i64 d = 0;
   i64 offset = 0;
-  i64 thunk_idx = 0;
+
+  // The smallest thunk index that is reachable from the current batch.
+  i64 t = 0;
 
   while (b < m.size()) {
-    // Move D foward as far as we can jump from B to anywhere in a thunk at D.
+    // Move D foward as far as we can jump from B to a thunk at D.
+    auto d_thunk_end = [&] {
+      u64 d_end = align_to(offset, 1 << m[d]->p2align) + m[d]->sh_size;
+      return align_to(d_end, Thunk::alignment) + max_thunk_size;
+    };
+
     while (d < m.size() &&
-           align_to(offset, 1 << m[d]->p2align) + m[d]->sh_size + max_thunk_size <
-           m[b]->offset + max_distance<E>()) {
+           (b == d || d_thunk_end() <= m[b]->offset + max_distance())) {
       offset = align_to(offset, 1 << m[d]->p2align);
       m[d]->offset = offset;
       offset += m[d]->sh_size;
@@ -239,34 +230,33 @@ void create_range_extension_thunks(Context<E> &ctx, OutputSection<E> &osec) {
     // to make sure that there's at least one section between B and C
     // to ensure progress.
     c = b + 1;
-    while (c < m.size() &&
-           m[c]->offset + m[c]->sh_size < m[b]->offset + batch_size<E>)
+    while (c < d && m[c]->offset + m[c]->sh_size < m[b]->offset + batch_size)
       c++;
 
     // Move A forward so that A is reachable from C.
     i64 c_offset = (c == m.size()) ? offset : m[c]->offset;
-    while (a < m.size() && m[a]->offset + max_distance<E>() < c_offset)
+    while (a < b && a < m.size() && m[a]->offset + max_distance() < c_offset)
       a++;
 
     // Erase references to out-of-range thunks.
-    while (thunk_idx < osec.thunks.size() &&
-           osec.thunks[thunk_idx]->offset < m[a]->offset)
-      reset_thunk(*osec.thunks[thunk_idx++]);
+    while (t < thunks.size() && thunks[t]->offset < m[a]->offset)
+      reset_thunk(*thunks[t++]);
 
-    // Create a thunk for input sections between B and C and place it at D.
-    offset = align_to(offset, RangeExtensionThunk<E>::alignment);
-    RangeExtensionThunk<E> *thunk =
-      new RangeExtensionThunk<E>(osec, osec.thunks.size(), offset);
-    osec.thunks.emplace_back(thunk);
+    // Create a new thunk and place it at D.
+    offset = align_to(offset, Thunk::alignment);
+    i64 thunk_idx = thunks.size();
+    Thunk *thunk = new Thunk(*this, offset);
+    thunks.emplace_back(thunk);
 
-    // Scan relocations between B and C to collect symbols that need thunks.
+    // Scan relocations between B and C to collect symbols that need
+    // entries in the new thunk.
     tbb::parallel_for_each(m.begin() + b, m.begin() + c,
                            [&](InputSection<E> *isec) {
-      scan_rels(ctx, *isec, *thunk);
+      scan_rels(ctx, *isec, *thunk, thunk_idx);
     });
 
     // Now that we know the number of symbols in the thunk, we can compute
-    // its size.
+    // the thunk's size.
     assert(thunk->size() < max_thunk_size);
     offset += thunk->size();
 
@@ -279,7 +269,7 @@ void create_range_extension_thunks(Context<E> &ctx, OutputSection<E> &osec) {
     // Assign offsets within the thunk to the symbols.
     for (i64 i = 0; i < thunk->symbols.size(); i++) {
       Symbol<E> &sym = *thunk->symbols[i];
-      sym.extra.thunk_idx = thunk->thunk_idx;
+      sym.extra.thunk_idx = thunk_idx;
       sym.extra.thunk_sym_idx = i;
     }
 
@@ -291,7 +281,7 @@ void create_range_extension_thunks(Context<E> &ctx, OutputSection<E> &osec) {
       std::span<RangeExtensionRef> range_extn = isec->extra.range_extn;
 
       for (i64 i = 0; i < rels.size(); i++)
-        if (range_extn[i].thunk_idx == thunk->thunk_idx)
+        if (range_extn[i].thunk_idx == thunk_idx)
           range_extn[i].sym_idx = syms[rels[i].r_sym]->extra.thunk_sym_idx;
     });
 
@@ -299,18 +289,16 @@ void create_range_extension_thunks(Context<E> &ctx, OutputSection<E> &osec) {
     b = c;
   }
 
-  while (thunk_idx < osec.thunks.size())
-    reset_thunk(*osec.thunks[thunk_idx++]);
+  while (t < thunks.size())
+    reset_thunk(*thunks[t++]);
 
-  osec.shdr.sh_size = offset;
+  this->shdr.sh_size = offset;
+
+  for (InputSection<E> *isec : members)
+    this->shdr.sh_addralign =
+      std::max<u32>(this->shdr.sh_addralign, 1 << isec->p2align);
 }
 
-using E = MOLD_TARGET;
-
-static_assert(max_thunk_size / E::thunk_size < INT16_MAX);
-
-template void create_range_extension_thunks(Context<E> &, OutputSection<E> &);
-
 } // namespace mold::elf
 
 #endif
